#include "QR.hh"
#include "alglin.hh"

namespace QR {

  // Purpose
  // =======
  // generates a real elementary reflector H of order n, such that
  //
  // H * ( x0 ) = ( beta ),   H^T * H = I.
  //     (  v )   (   0  )
  //
  // where alpha and beta are scalars, beta is non-negative, and x is
  // an (n-1)-element real vector.  H is represented in the form
  //
  //                     2
  //     H = I - ---------------  ( alpha ) * ( alpha v' ),
  //              alpha^2 + v^2   ( v     )
  //
  //                                  2
  //     H * ( x0 ) = ( x0 ) - -------------- ( alpha ) * ( alpha v' ) ( x0 ) 
  //         (  v )   (  v )    alpha^2+v^2   (   v   )                (  v )
  //
  //                            2*(x0*alpha+v^2)
  //                = ( x0 ) - ------------------ ( alpha ) 
  //                  (  v )     alpha^2+v^2      ( v    )
  //
  //    alpha = sign(x0) * (|x0| + || (x0,xv) ||)
  //
  //                  2 alpha * ( alpha x0 + v^2)
  //    beta = x0 - ------------------------------
  //                        alpha^2 + v^2  
  //
  //    H = I - gamma ( 1       ) * ( 1 v'/alpha ),
  //                  ( v/alpha )
  //
  //    gamma = 2*alpha^2/(alpha^2+v'*v) 
  //

  valueType // The value tau.
  buildReflection( indexType n,        // The order of the elementary reflector.
                   valueType x[],      // On entry, the vector x. On exit, it is overwritten with the vector v and beta
                   valueType xNorm ) { // The norm of vector x
     // xNorm^2 = x0^2 + v^2 
     valueType x0     = x[0] ;
     valueType xNorm2 = xNorm*xNorm ;
     valueType vNorm2 = xNorm2-x0*x0 ;
     valueType alpha  = ( x0 > 0 ? x0 + xNorm : x0 - xNorm ) ;
     valueType gamma  = 2/(1+vNorm2/(alpha*alpha)) ;
     x[0] -= gamma * ( x0 + vNorm2 / alpha ) ; // beta
     alglin::scal( n-1, 1/alpha, x+1, 1 ) ;
     return gamma ;
  }

  indexType
  factorize( indexType const n,      /* Number of rows of matrix A             */
             indexType const m,      /* Number of columns of matrix A          */
             valueType       A[],    /* Pointer to coefficients of A           */
             indexType const lda,    /* Leading dimension of A, Fortran style  */
             indexType       ipiv[], /* Permutation vector                     */
             valueType       work[] ) {
    indexType nm = m < n ? m : n-1 ;
    for ( indexType i = 0 ; i < nm ; ++i ) {
      // evaluate column norm
      for ( indexType j = i ; j < m ; ++j ) work[j] = alglin::nrm2( n-i, A+i+j*lda, 1 ) ;
      // find the column with maximum norm
      indexType ip = i+alglin::iamax( m-i, work+i, 1 ) ;
      if ( work[ip] == 0 ) return i ; // no more column found
      // swap columns
      if ( i != ip ) alglin::swap( n, A + i*lda, 1, A + ip*lda, 1 ) ;
      ipiv[i] = ip ; // store the pivot index to vector ipiv
      // build householder reflection
      valueType * Ai = A + i*(lda+1) ;
      work[i] = buildReflection( n-i, Ai, work[ip] ) ;
      // apply householder reflection to submatrix
      valueType * w = work + i + 1 ;
      // Ai = ( beta v )^T not (1 v^T)
      valueType tmp = *Ai ;
      *Ai = 1 ;
      // gemv: w <- beta * w + alpha * A^T x
      //                         row  column alpha  A            x      beta w
      alglin::gemv( "Transpose", n-i, m-i-1, 1.0,   Ai+lda, lda, Ai, 1, 0.0, w, 1 ) ;
      // ger:  A <- A + alpha u v^T
      //           row  column alpha     A      u     v
      alglin::ger( n-i, m-i-1, -work[i], Ai, 1, w, 1, Ai+lda, lda ) ;
      *Ai = tmp ;
    }
    return 0 ;
  }

  void
  applyReflections( indexType const n,      /* Number of rows of matrix QR             */
                    indexType const m,      /* Number of columns of matrix QR          */
                    valueType const QR[],   /* Pointer to coefficients of QR           */
                    indexType const lda,    /* Leading dimension of QR, Fortran style  */
                    indexType const ipiv[], /* Permutation vector                      */
                    valueType const work[],
                    valueType       v[],
                    indexType       incV ) {
    indexType nm = m < n ? m : n-1 ;
    for ( indexType i = 0 ; i < nm ; ++i ) {
      valueType       * vi = v  + i*incV ;
      valueType const * Qi = QR + i*lda + i + 1 ;
      valueType bf = work[i] * (*vi + alglin::dot( n-i-1, Qi, 1, vi + incV, incV )) ;
      *vi -= bf ;
      alglin::axpy( n-i-1, -bf, Qi, 1, vi + incV, 1 ) ;
    }
  }

  void
  solve( indexType const n,      /* Number of rows and columns of matrix A */
         valueType const QR[],   /* Pointer to coefficients of LU          */
         indexType const lda,    /* Leading dimension of LU, Fortran style */
         indexType const ipiv[], /* Permutation vector                     */
         valueType const work[],
         valueType       b[],    /* Pointer to coefficients of vector b    */
         indexType       incB ) {
    applyReflections( n, n, QR, lda, ipiv, work, b, incB ) ;
    alglin::trsv( "Upper", "No Transpose", "Non Unit", n, QR, lda, b, incB ) ;
    for ( indexType i = 0 ; i < n-1 ; ++i ) {
      indexType ip = ipiv[i] ;
      if ( ip != i ) std::swap( b[i*incB], b[ip*incB] ) ;
    }

  }
}
